/*
 * Copyright (C) 1999 Yasuhiro Ohara
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

#include "ospf6d.h"

#include "linklist.h"
#include "ospf6_lsa.h"

void
ospf6_route_hook_call (int type, struct ospf6_lsa *lsa,
                       struct ospf6_lsdb *lsdb);

struct ospf6_lsdb *
ospf6_lsdb_create ()
{
  struct ospf6_lsdb *lsdb;

  lsdb = XCALLOC (MTYPE_OSPF6_LSDB, sizeof (struct ospf6_lsdb));
  if (lsdb == NULL)
    {
      zlog_warn ("Can't malloc lsdb");
      return NULL;
    }

  lsdb->hook_list[0] = linklist_create ();
  lsdb->hook_list[1] = linklist_create ();
  lsdb->hook_list[2] = linklist_create ();

  return lsdb;
}

void
ospf6_lsdb_delete (struct ospf6_lsdb *lsdb)
{
  ospf6_lsdb_remove_all (lsdb);
  linklist_delete (lsdb->hook_list[0]);
  linklist_delete (lsdb->hook_list[1]);
  linklist_delete (lsdb->hook_list[2]);
  XFREE (MTYPE_OSPF6_LSDB, lsdb);
}

void
ospf6_lsdb_remove_maxage (struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;
  struct ospf6_lsa *lsa;

  for (node = ospf6_lsdb_head (lsdb); node; node = ospf6_lsdb_next (node))
    {
      lsa = node->lsa;

      /* contiue if it's not MaxAge */
      if (! IS_LSA_MAXAGE (lsa))
        continue;

      /* continue if it's referenced by some retrans-lists */
      if (lsa->lock != 1)
        continue;

      if (IS_OSPF6_DUMP_LSDB)
        zlog_info ("Remove MaxAge LSA: %s", lsa->str);

#if 0
      /* hooks should had been called already.
         they should be called at expire time */
      CALL_LSA_DATABASE_REMOVE (lsa->header->type, lsa);
      CALL_REMOVE_HOOK (&database_hook, lsa);

      if (lsdb->hook)
        (*lsdb->hook) (lsa, NULL);
#endif

      ospf6_lsdb_remove (lsa, lsdb);
    }
}

struct ospf6_lsa *
ospf6_lsdb_lookup (u_int16_t type, u_int32_t id, u_int32_t adv_router,
                   void *scope)
{
  struct ospf6_interface *o6i;
  struct ospf6_area *o6a;
  listnode i, j;

#if 0
  assert (scope);
#else /*0*/
  if (! scope)
    {
      char type_buf[32], id_buf[32], adv_router_buf[32];
      ospf6_lsa_type_string (type, type_buf, sizeof (type_buf));
      inet_ntop (AF_INET, &id, id_buf, sizeof (id_buf));
      inet_ntop (AF_INET, &adv_router, adv_router_buf,
                 sizeof (adv_router_buf));
      zlog_warn ("Looking for %s id=%s adv_router=%s: Unknown scope",
                 type_buf, id_buf, adv_router_buf);
    }
#endif /*0*/

  if (scope == (void *) ospf6)
    return ospf6_lsdb_lookup_lsdb (type, id, adv_router, ospf6->lsdb);

  for (i = listhead (ospf6->area_list); i; nextnode (i))
    {
      o6a = getdata (i);

      if (scope == (void *) o6a)
        return ospf6_lsdb_lookup_lsdb (type, id, adv_router, o6a->lsdb);

      for (j = listhead (o6a->if_list); j; nextnode (j))
        {
          o6i = getdata (j);

          if (scope == (void *) o6i)
            return ospf6_lsdb_lookup_lsdb (type, id, adv_router, o6i->lsdb);
        }
    }

  zlog_warn ("Can't lookup lsdb: type %#x: unknown scope", ntohs (type));
  return NULL;
}

void
ospf6_lsdb_install (struct ospf6_lsa *new)
{
  struct ospf6_lsdb *lsdb;
  struct ospf6_lsa *old;
  int recalc = 0;

  struct ospf6 *as = NULL;
  struct ospf6_area *area = NULL;
  struct ospf6_interface *linklocal = NULL;

  if (OSPF6_LSA_IS_SCOPE_LINKLOCAL (ntohs (new->header->type)))
    {
      linklocal = (struct ospf6_interface *) new->scope;
      lsdb = linklocal->lsdb;
    }
  else if (OSPF6_LSA_IS_SCOPE_AREA (ntohs (new->header->type)))
    {
      area = (struct ospf6_area *) new->scope;
      lsdb = area->lsdb;
    }
  else if (OSPF6_LSA_IS_SCOPE_AS (ntohs (new->header->type)))
    {
      as = (struct ospf6 *) new->scope;
      lsdb = as->lsdb;
    }
  else
    {
      zlog_err ("Can't install lsdb: scope unknown: %s", new->str);
      return;
    }

  /* whether schedule calculation or not */
  old = ospf6_lsdb_lookup_lsdb (new->header->type, new->header->id,
                                new->header->adv_router, lsdb);
  if (! old || ospf6_lsa_differ (old, new))
    recalc = 1;

  /* log */
  if (IS_OSPF6_DUMP_LSDB)
    zlog_info ("LSDB: Install: %s %s", new->str,
               ((IS_LSA_MAXAGE (new)) ? "(MaxAge)" : ""));

  ospf6_lsdb_add (new, lsdb);
  gettimeofday (&new->installed, NULL);

#if 0
  if (lsdb->hook)
    (*lsdb->hook) (old, new);
#else /*0*/
  if (! old)
    {
      CALL_ADD_HOOK (&database_hook, new);
    }
  else if (old && recalc)
    {
      CALL_CHANGE_HOOK (&database_hook, new);
    }
#endif /*0*/

  if (old)
    ospf6_lsdb_remove (old, lsdb);
}



/* vty functions */

#define OSPF6_LSDB_MATCH_TYPE        0x01
#define OSPF6_LSDB_MATCH_ID          0x02
#define OSPF6_LSDB_MATCH_ADV_ROUTER  0x04
#define OSPF6_LSDB_SHOW_DUMP         0x08
#define OSPF6_LSDB_SHOW_SUMMARY      0x10

static int
ospf6_lsdb_match (int flag, u_int16_t type, u_int32_t id,
                  u_int32_t adv_router, struct ospf6_lsa *lsa)
{
  if (CHECK_FLAG (flag, OSPF6_LSDB_MATCH_TYPE) &&
      lsa->header->type != type)
    return 0;

  if (CHECK_FLAG (flag, OSPF6_LSDB_MATCH_ID) &&
      lsa->header->id != id)
    return 0;

  if (CHECK_FLAG (flag, OSPF6_LSDB_MATCH_ADV_ROUTER) &&
      lsa->header->adv_router != adv_router)
    return 0;

  return 1;
}

int
show_ipv6_ospf6_lsdb (struct vty *vty, int argc, char **argv,
                      struct ospf6_lsdb *lsdb)
{
  u_int flag;
  u_int16_t type = 0;
  u_int32_t id, adv_router;
  int ret;
  struct ospf6_lsdb_node *node;
  char invalid[32], *invalidp;
  int l_argc = argc;
  char **l_argv = argv;

  flag = 0;
  memset (invalid, 0, sizeof (invalid));
  invalidp = invalid;

  /* chop tail if the words is 'dump' or 'summary' */
  if (l_argc > 0 && ! strcmp (l_argv[l_argc - 1], "dump"))
    {
      SET_FLAG (flag, OSPF6_LSDB_SHOW_DUMP);
      l_argc --;
    }
  else if (l_argc > 0 && ! strcmp (l_argv[l_argc - 1], "summary"))
    {
      SET_FLAG (flag, OSPF6_LSDB_SHOW_SUMMARY);
      l_argc --;
    }

  if (l_argc > 0)
    {
      SET_FLAG (flag, OSPF6_LSDB_MATCH_TYPE);
      if (! strncmp (l_argv[0], "r", 1))
        type = htons (OSPF6_LSA_TYPE_ROUTER);
      if (! strncmp (l_argv[0], "n", 1))
        type = htons (OSPF6_LSA_TYPE_NETWORK);
      if (! strncmp (l_argv[0], "a", 1))
        type = htons (OSPF6_LSA_TYPE_AS_EXTERNAL);
      if (! strcmp (l_argv[0], "intra-prefix"))
        type = htons (OSPF6_LSA_TYPE_INTRA_PREFIX);
      if (! strcmp (l_argv[0], "inter-router"))
        type = htons (OSPF6_LSA_TYPE_INTER_ROUTER);
      if (! strcmp (l_argv[0], "inter-prefix"))
        type = htons (OSPF6_LSA_TYPE_INTER_PREFIX);
      if (! strncmp (l_argv[0], "l", 1))
        type = htons (OSPF6_LSA_TYPE_LINK);
      if (! strncmp (l_argv[0], "0x", 2) && strlen (l_argv[0]) == 6)
        type = htons ((short) strtol (l_argv[0], (char **)NULL, 16));
      if (! strncmp (l_argv[0], "*", 1))
        UNSET_FLAG (flag, OSPF6_LSDB_MATCH_TYPE);
    }

  if (l_argc > 1)
    {
      SET_FLAG (flag, OSPF6_LSDB_MATCH_ID);
      if (! strncmp (l_argv[1], "*", 1))
        UNSET_FLAG (flag, OSPF6_LSDB_MATCH_ID);
      else
        {
          ret = inet_pton (AF_INET, l_argv[1], &id);
          if (ret != 1)
            {
              id = htonl (strtoul (l_argv[1], &invalidp, 10));
              if (invalid[0] != '\0')
                {
                  vty_out (vty, "Link State ID is not parsable: %s%s",
                           l_argv[1], VTY_NEWLINE);
                  return CMD_SUCCESS;
                }
            }
        }
    }

  if (l_argc > 2)
    {
      SET_FLAG (flag, OSPF6_LSDB_MATCH_ADV_ROUTER);
      if (! strncmp (l_argv[2], "*", 1))
        UNSET_FLAG (flag, OSPF6_LSDB_MATCH_ADV_ROUTER);
      else
        {
          ret = inet_pton (AF_INET, l_argv[2], &adv_router);
          if (ret != 1)
            {
              adv_router = htonl (strtoul (l_argv[2], &invalidp, 10));
              if (invalid[0] != '\0')
                {
                  vty_out (vty, "Advertising Router is not parsable: %s%s",
                           l_argv[2], VTY_NEWLINE);
                  return CMD_SUCCESS;
                }
            }
        }
    }

  if (CHECK_FLAG (flag, OSPF6_LSDB_SHOW_SUMMARY))
    ospf6_lsa_show_summary_header (vty);

  for (node = ospf6_lsdb_head (lsdb); node;
       node = ospf6_lsdb_next (node))
    {
      if (! ospf6_lsdb_match (flag, type, id, adv_router, node->lsa))
        continue;

      if (CHECK_FLAG (flag, OSPF6_LSDB_SHOW_DUMP))
        ospf6_lsa_show_dump (vty, node->lsa);
      else if (CHECK_FLAG (flag, OSPF6_LSDB_SHOW_SUMMARY))
        ospf6_lsa_show_summary (vty, node->lsa);
      else
        ospf6_lsa_show (vty, node->lsa);
    }

  return CMD_SUCCESS;
}

DEFUN (show_ipv6_ospf6_database,
       show_ipv6_ospf6_database_cmd,
       "show ipv6 ospf6 database",
       SHOW_STR
       IP6_STR
       OSPF6_STR
       "LSA Database\n"
       )
{
  struct ospf6_area *o6a;
  struct ospf6_interface *o6i;
  listnode i, j;

  /* call show function for each of LSAs in the LSDBs */

  for (i = listhead (ospf6->area_list); i; nextnode (i))
    {
      o6a = (struct ospf6_area *) getdata (i);

      /* LinkLocal LSDBs */
      for (j = listhead (o6a->if_list); j; nextnode (j))
        {
          o6i = (struct ospf6_interface *) getdata (j);

          vty_out (vty, "%s", VTY_NEWLINE);
          vty_out (vty, "                Interface %s (Area: %s):%s",
                   o6i->interface->name, o6a->str, VTY_NEWLINE);
          vty_out (vty, "%s", VTY_NEWLINE);
          show_ipv6_ospf6_lsdb (vty, argc, argv, o6i->lsdb);
        }

      /* Area LSDBs */
      vty_out (vty, "%s", VTY_NEWLINE);
      vty_out (vty, "                Area %s:%s", o6a->str, VTY_NEWLINE);
      vty_out (vty, "%s", VTY_NEWLINE);
      show_ipv6_ospf6_lsdb (vty, argc, argv, o6a->lsdb);
    }

  /* AS LSDBs */
  vty_out (vty, "%s", VTY_NEWLINE);
  vty_out (vty, "                AS:%s", VTY_NEWLINE);
  vty_out (vty, "%s", VTY_NEWLINE);
  show_ipv6_ospf6_lsdb (vty, argc, argv, ospf6->lsdb);

  return CMD_SUCCESS;
}

ALIAS (show_ipv6_ospf6_database,
       show_ipv6_ospf6_database_type_cmd,
       "show ipv6 ospf6 database (router|network|as-external|intra-prefix|inter-prefix|inter-router|link|*|dump|summary)",
       SHOW_STR
       IP6_STR
       OSPF6_STR
       "LSA Database\n"
       "Router-LSA\n"
       "Network-LSA\n"
       "AS-External-LSA\n"
       "Intra-Area-Prefix-LSA\n"
       "Inter-Area-Router-LSA\n"
       "Inter-Area-Prefix-LSA\n"
       "Link-LSA\n"
       "All LS Type\n"
       "Dump raw LSA data in Hex\n"
       "show summary of LSA\n"
       )

ALIAS (show_ipv6_ospf6_database,
       show_ipv6_ospf6_database_type_id_cmd,
       "show ipv6 ospf6 database (router|network|as-external|intra-prefix|inter-prefix|inter-router|link|*) (A.B.C.D|*|dump|summary)",
       SHOW_STR
       IP6_STR
       OSPF6_STR
       "LSA Database\n"
       "Router-LSA\n"
       "Network-LSA\n"
       "AS-External-LSA\n"
       "Intra-Area-Prefix-LSA\n"
       "Inter-Area-Router-LSA\n"
       "Inter-Area-Prefix-LSA\n"
       "Link-LSA\n"
       "All LS Type\n"
       "Link State ID\n"
       "All Link State ID\n"
       "Dump raw LSA data in Hex\n"
       "show summary of LSA\n"
       )

ALIAS (show_ipv6_ospf6_database,
       show_ipv6_ospf6_database_type_id_adv_router_cmd,
       "show ipv6 ospf6 database (router|network|as-external|intra-prefix|inter-prefix|inter-router|link|*) (A.B.C.D|*) (A.B.C.D|*|dump|summary)",
       SHOW_STR
       IP6_STR
       OSPF6_STR
       "LSA Database\n"
       "Router-LSA\n"
       "Network-LSA\n"
       "AS-External-LSA\n"
       "Intra-Area-Prefix-LSA\n"
       "Inter-Area-Router-LSA\n"
       "Inter-Area-Prefix-LSA\n"
       "Link-LSA\n"
       "All LS Type\n"
       "Link State ID\n"
       "All Link State ID\n"
       "Advertising Router\n"
       "All Advertising Router\n"
       "Dump raw LSA data in Hex\n"
       "show summary of LSA\n"
       )

ALIAS (show_ipv6_ospf6_database,
       show_ipv6_ospf6_database_type_id_adv_router_dump_cmd,
       "show ipv6 ospf6 database (router|network|as-external|intra-prefix|inter-prefix|inter-router|link|*) (A.B.C.D|*) (A.B.C.D|*) (dump|summary|)",
       SHOW_STR
       IP6_STR
       OSPF6_STR
       "LSA Database\n"
       "Router-LSA\n"
       "Network-LSA\n"
       "AS-External-LSA\n"
       "Intra-Area-Prefix-LSA\n"
       "Inter-Area-Router-LSA\n"
       "Inter-Area-Prefix-LSA\n"
       "Link-LSA\n"
       "All LS Type\n"
       "Link State ID\n"
       "All Link State ID\n"
       "Advertising Router\n"
       "All Advertising Router\n"
       "Dump raw LSA data in Hex\n"
       "show summary of LSA\n"
       )

void
ospf6_lsdb_init ()
{
  install_element (VIEW_NODE, &show_ipv6_ospf6_database_cmd);
  install_element (VIEW_NODE, &show_ipv6_ospf6_database_type_cmd);
  install_element (VIEW_NODE, &show_ipv6_ospf6_database_type_id_cmd);
  install_element (VIEW_NODE, &show_ipv6_ospf6_database_type_id_adv_router_cmd);
  install_element (VIEW_NODE, &show_ipv6_ospf6_database_type_id_adv_router_dump_cmd);

  install_element (ENABLE_NODE, &show_ipv6_ospf6_database_cmd);
  install_element (ENABLE_NODE, &show_ipv6_ospf6_database_type_cmd);
  install_element (ENABLE_NODE, &show_ipv6_ospf6_database_type_id_cmd);
  install_element (ENABLE_NODE, &show_ipv6_ospf6_database_type_id_adv_router_cmd);
  install_element (ENABLE_NODE, &show_ipv6_ospf6_database_type_id_adv_router_dump_cmd);
}

/***/

static struct ospf6_lsdb_node *
ospf6_lsdb_node_create ()
{
  struct ospf6_lsdb_node *node;

  node = XMALLOC (MTYPE_OSPF6_LSDB, sizeof (struct ospf6_lsdb_node));
  memset (node, 0, sizeof (struct ospf6_lsdb_node));

  return node;
}

static void
ospf6_lsdb_node_delete (struct ospf6_lsdb_node *node)
{
  XFREE (MTYPE_OSPF6_LSDB, node);
}

static void
ospf6_lsdb_node_lock (struct ospf6_lsdb_node *node)
{
  node->lock ++;
}

static void
ospf6_lsdb_node_unlock (struct ospf6_lsdb_node *node)
{
  node->lock --;
  if (node->lock == 0)
    {
      /* list manipulation */
      if (node->prev)
        node->prev->next = node->next;
      else
        node->lsdb->head = node->next;

      if (node->next)
        node->next->prev = node->prev;
      else
        node->lsdb->tail = node->prev;

      node->lsdb->count --;
      ospf6_lsdb_node_delete (node);
    }
}

struct ospf6_lsdb_node *
ospf6_lsdb_head (struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;

  node = lsdb->head;
  if (node)
    ospf6_lsdb_node_lock (node);

  return node;
}

struct ospf6_lsdb_node *
ospf6_lsdb_next (struct ospf6_lsdb_node *node)
{
  struct ospf6_lsdb_node *next;

  next = node->next;
  if (next)
    ospf6_lsdb_node_lock (next);

  ospf6_lsdb_node_unlock (node);
  return next;
}

/* call this function when 'break'ing 'for' loop */
void
ospf6_lsdb_end (struct ospf6_lsdb_node *node)
{
  ospf6_lsdb_node_unlock (node);
}

static struct ospf6_lsdb_node *
ospf6_lsdb_node_lookup (u_int16_t type, u_int32_t id, u_int32_t adv_router,
                        struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;

  for (node = ospf6_lsdb_head (lsdb); node; node = ospf6_lsdb_next (node))
    {
      if (node->lsa->header->type != type)
        continue;
      if (node->lsa->header->id != id)
        continue;
      if (node->lsa->header->adv_router != adv_router)
        continue;

      ospf6_lsdb_end (node);
      return node;
    }
  return NULL;
}

#define ADD    0
#define CHANGE 1
#define REMOVE 2

void
ospf6_lsdb_add (struct ospf6_lsa *lsa, struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;

  /* create lsdb node */
  node = ospf6_lsdb_node_create ();
  node->lsdb = lsdb;

  /* attach LSA to the node */
  node->lsa = lsa;
  ospf6_lsa_lock (node->lsa);

  /* list manipulation */
  node->prev = lsdb->tail;
  if (lsdb->head == NULL)
    lsdb->head = node;
  else
    lsdb->tail->next = node;
  lsdb->tail = node;

  ospf6_lsdb_node_lock (node);
  lsdb->count ++;
}

void
ospf6_lsdb_remove (struct ospf6_lsa *lsa, struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;

  for (node = ospf6_lsdb_head (lsdb); node; node = ospf6_lsdb_next (node))
    {
      if (node->lsa == lsa)
        {
          ospf6_lsa_unlock (node->lsa);
          ospf6_lsdb_node_unlock (node);
        }
    }
}

void
ospf6_lsdb_remove_all (struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;
  for (node = ospf6_lsdb_head (lsdb); node; node = ospf6_lsdb_next (node))
    ospf6_lsdb_remove (node->lsa, lsdb);
}

struct ospf6_lsa *
ospf6_lsdb_lookup_lsdb (u_int16_t type, u_int32_t id, u_int32_t adv_router,
                        struct ospf6_lsdb *lsdb)
{
  struct ospf6_lsdb_node *node;
  node = ospf6_lsdb_node_lookup (type, id, adv_router, lsdb);
  if (node)
    return node->lsa;
  return NULL;
}

